/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


/*
TODO:
The letters of the non-standard comm. commands are a bit too random,
fix it before it is too late!
*/


/*
  This file deserves a little explanation!

  Instead of creating a huge circular buffer for uart data,
  a smaller circular buffer 'command_queue' are used instead.
  This buffer contains the address and arguments to functions which
  can fetch bytes from any place in the memory.

  eg. MegaTune sends an 'A' to request realtime variables:
  1) Uart interrupt calls comm()
  2) comm pushes the address of sendRTvar and the number of
     bytes to send (22) onto the command queue and enables tx-interrupt.
  3) tx-interrupt reads the command queue, executes the function and
     decrements the number of bytes to send (and stores it in the
     command queue again). If number equals zero remove the function
     from the queue and start over. If the queue is empty, then disable
     interrupt.
*/

   /*!
   * \addtogroup g_serial_comm Serial Communication Routines
   */
/*@{*/


#include "global.h"
#include "comm.h"
#include "aim.h"
#include "ringbuf.h"
#include "storage.h"
#include "actuators.h"
#include "ve.h"
#include "iac.h"
#include "adc.h"
#include "lcd.h"
#include "ego.h"
#include "fuelcalc.h"
#include "helpers.h"
#include "benchmark.h"
#include "wbo2.h"
#include "boostcontrol.h"
#include "slcan.h"
#include "kb.h"
#include "spi.h"
#include "firmware_revision.h"
#include "knock.h"
#include "fuelcontrol.h"
#include "my_make.h"

#include "triggerframe.h"

#ifdef MAINLOOP_TASKER
#include "mainloop.h"
#endif

#ifdef A340E
#include "a340e.h"
#endif

#include <avr/io.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include <avr/sfr_defs.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "varstr.h"
#include "varstr_ext.h"
#include "varstr_shared.h"

#include "arithm.c"
#include <util/crc16.h>

struct cmd_queue_t cmd_queue[CMD_QUEUE_SIZE];
uint8_t cmd_queue_tail, cmd_queue_head;

struct com_t com;

char vems_string[] PROGMEM = "VEMS Genboard v3, ";

/* baudrate change pending flag. Ignored when 0,
 * otherwise it contains the new baudrate divisor */
uint8_t new_baudrate;

// TABLES_TO_DUMP1 lives in menu.c because otherwise gcc fucks up flashsize by +500 bytes beyond imagination
uint8_t PROGMEM TABLES_TO_DUMP2[] = {
'n',  2+IGNADV_SIZE_KPA-1

#ifdef HWMAP
,'h', HWMAP_LINES-1
#endif

,'i', 2+ MAP_CORR_SIZE_KPA-1
,'w', 2+ LAMBDA_SIZE_KPA-1
,'s', 2-1
,'p', 3-1 // 3 anytrim1_array lines (17 bytes)
};

 /** send echo via UART0,
 *allow bytes coming from UART1 within next 2.1 seconds to be relayed back to UART0
 *after 2.1sec elapsed, 1wire status ,msB and ssB send on UART0
 */
void status_com1()
{
    ringbuf_push(&comm_sndbuf, '-' );
	pushfuncpar( &srl_direct, 4, (void*)( ('+' << 8) | 'u' ) ); // will print "+uzz"

	UCSR0B |= _BV(UDRIE0); // enable transmit interrupt for UART0, so comm_sndbuf does not get stuck
	// allow 2.1 sec window for bytes coming from UART1 to be transmitted (relayed) via UART0:
	com.comm1_relay_allowed = 100; // 100 * 0.1 seconds

}

#ifdef ONE_WIRE
//http://www.vems.hu/wiki/index.php?page=ButtonImmobilizer
struct one_wire_t one_wire;

#define oneWireInitLen 14
const uint8_t oneWireInit[oneWireInitLen] PROGMEM = {0xc1,
0xc1, 0xe1, 0x33, 0xff,  0xff, 0xff, 0xff, 0xff,  0xff, 0xff, 0xff, 0xe3, 0xc1 };

#ifdef ONE_WIRE_STAY_ALIVE
static inline void one_wire_adc_charge(void)// set one_wire adc output to high, to charge the capacitor
{
    if ( ( config_s.prohibit_config & _BV(PROHIBIT_CONF_IBUTTON_KEEPALIVE)) == 0 ) return;// ibutton not configured to adc keepalive
    PORTF |= _BV(6); //only use the integrated pullup to charge
	//DDRF |= _BV(6);
}

void check_1wire_adc(void)
{
    if ( ( config_s.prohibit_config & _BV(PROHIBIT_CONF_IBUTTON_KEEPALIVE)) == 0 ) return;// ibutton not configured to adc keepalive
    one_wire.status = ONE_WIRE_ADC_CHECK;
    mainlp.adc_please|=_BV(ONE_WIRE_ADC);
}
#endif // ONE_WIRE_STAY_ALIVE

/** send 1-wire magic sequence via UART1
*/
void read_1wire()
{
    if ( ( config_s.prohibit_config & _BV(PROHIBIT_CONF_IBUTTON)) == 0 ) return;// ibutton not configured
    if (one_wire.status == ONE_WIRE_FOUND_CORRECT) return; //ibutton found and correct
#ifdef ONE_WIRE_STAY_ALIVE
    if (one_wire.status == ONE_WIRE_ADC_CHECK){//check adc input for keepalive
        if (mainlp.adc_please &_BV(ONE_WIRE_ADC)) return; //reading adc not finished
        if (sensors8(ONE_WIRE_ADC) > ONE_WIRE_ADC_LIMIT){
            one_wire_adc_charge();
            one_wire.status = ONE_WIRE_FOUND_CORRECT;

        }else{
            one_wire.status = ONE_WIRE_LISTENING;
        }
        return;//1-wire interface not checking while the adc not measured
    }
#endif // ONE_WIRE_STAY_ALIVE

    if (one_wire.status == ONE_WIRE_CRC_OK){//1wire read success, crc ok, test if code correct
        if ( ( config_s.prohibit_h ==one_wire.ssB)
          && ( config_s.prohibit_l ==one_wire.msB)){//code correct
#ifdef ONE_WIRE_STAY_ALIVE
              one_wire_adc_charge();
#endif // ONE_WIRE_STAY_ALIVE
              one_wire.status = ONE_WIRE_FOUND_CORRECT;
          }else{
              one_wire.status = ONE_WIRE_FOUND_INCORRECT;//crc ok, but code incorrect, need it for testing ibutton and show it
          }
          return;
    }

    if (one_wire.tx_cnt !=0) return;
    one_wire.status = ONE_WIRE_LISTENING;
    one_wire.tx_cnt=1;
    one_wire.rx_cnt=0;
    UDR1 = pgm_read_byte( &(oneWireInit[0] ));//send the first byte
    UCSR1B |=_BV(UDRE1);// enable data register empty interrupt on usart1

/* old busy-wait based method
	for(i=0; i< oneWireInitLen; i++){
		while( (UCSR1A&0x20) == 0 )  ;  // UDRE1 bit means: TX-Data Register Empty

		UDR1 = pgm_read_byte( &(oneWireInit[i] ));
	}
*/
}

#endif //ONE_WIRE

/***************************************************************************/
/***************************************************************************/
/*** MegaTune compatibility functions                                    ***/
/***************************************************************************/
/***************************************************************************/

uint8_t sendSignature(uint8_t i) {
  if (i < 18) { // TODO: check sizeof ?
        return pgm_read_byte( (uint8_t *)&vems_string + i );
  }
  i -= 18;
  if (i < 6)
        return *( (uint8_t *)&FIRMWARE_VERSION_STR + i );
  switch (i)
  {
  case 7:	// send last
	  return MY_MAKE_BYTE_0;
  case 6:	// send first
	  return MY_MAKE_BYTE_1;
  default:
	  return 0;
  }
}

/***************************************************************************/
/*** codeVer                                                             ***/
/***************************************************************************/
// return codeversion "VEMS v1.1 12x12 kpa=2,1"
uint8_t MTcodeVer(uint8_t i) {
  switch(i){
  	case 0: return 'V';
  	case 1: return 'E';
  	case 2: return 'M';
  	case 3: return 'S';
  	case 4: return ' ';
  	case 5: return 'v';
  	case 6: return '0' + FIRMWARE_VERSION_MAJOR;
  	case 7: return '.';
  	case 8: return '0' + FIRMWARE_VERSION_MINOR;
  	case 9: return '.';
  	case 10: return '0' + FIRMWARE_VERSION_BUILD / 10; // TODO: make sure gcc optimizes this constant
  	case 11: return '0' + FIRMWARE_VERSION_BUILD % 10;
  	case 12: return ' ';
  	case 13: return '0' + VE_SIZE_RPM / 10; // TODO: make sure gcc optimizes this constant
  	case 14: return '0' + VE_SIZE_RPM % 10;
  	case 15: return 'x';
  	case 16: return '0' + VE_SIZE_KPA / 10; // TODO: make sure gcc optimizes this constant
  	case 17: return '0' + VE_SIZE_KPA % 10;
  	case 18: return ' ';
  	case 19: return 'k';
  	case 20: return 'p';
  	case 21: return 'a';
  	case 22: return '=';
  	case 23: return KPAFAC_UNIT == 0xF ? '2' : hex_hi(config.kpa_conf);
  	case 24: return ',';
  	case 25: return KPA_CONFIG_RESOLUTION == 0xF ? '1' : hex_lo(config.kpa_conf);
	default: return 'V';
    }
#if(MT_CODEVER_LEN != 26)
#warning "MT_CODEVER_LEN size mismatch"
#endif

}

/***************************************************************************/
/*** testComm                                                            ***/
/***************************************************************************/
// return low part of second-count
uint8_t MTtestComm(uint8_t i) {
  return (rtc.sec & 0xFF);
}



/***************************************************************************/
/*** getMCPdataHIGH                                                      ***/
/***************************************************************************/
// does not start a adc conversion, just masks out the high byte of latest conversion
// and returns if the channel data is not 0xffff

// channels range from 0-7 internally (from 1 to 8 according to datasheet)

uint8_t getMCPdataHIGH(uint8_t i) {

	uint16_t t;
	uint8_t r = 0;

	if(i < 8) {
		t = mcp3208_data[i];
		if (t != 0xffff)
	//		r = (uint8_t) t >> 8;
			r = t >> 8;
	}

	return r;
}


/***************************************************************************/
/*** getMCPdataLOW                                                       ***/
/***************************************************************************/
// starts an mpc channel adc conversion on the appropriate channel
// and if conversion is finished return low byte of adc result, else 0

// channels range from 0-7 internally (from 1 to 8 according to datasheet)

uint8_t getMCPdataLOW(uint8_t i) {

	uint16_t t;
	uint8_t r = 0;

	if(i < 8) {
		t = mcp3208_data[i];
		if (t != 0xffff)
			r = (uint8_t) t & 0xFF;
#ifndef MAINLOOP_TASKER //not needed for tasker
		read_mcp3208(i);
#endif
	}

	return r;
}

/*
* Sends the data with byte stuffing, but without heading
* needed for config table stuffed dump
* SendPage_stuffed must call after this
*/
uint8_t SendPage_data_stuffed (uint8_t cnt)
{
	//cmdqueue[head].private must have pageID
    ringbuf_push_bytestuff_updatecrc(&comm_sndbuf,send_page(cnt));
    return 'D';

}
/* sends the selected page in triggerlog format, with address or page name
format: N byte data + 2 byte address ( or 0xff+page ID) + 1byte SEND_PAGE_STUFFED_TYPE(119) + 2 byte CRC + HDLC_END
*/
uint8_t SendPage_stuffed( uint8_t i )
{
    uint8_t size = cmd_queue[cmd_queue_head].len -1 ;
    uint8_t revese_cnt= size-i;

	uint8_t v;
	if(i>=(size)){
		ringbuf_push(&comm_sndbuf, HDLC_END); /* send framing char, do NOT stuff*/
		com.commcrc = 0xffff;

	} else if(revese_cnt>=3){/* data section with address and trtype*/

        uint16_t addr_or_page = (uint16_t)cmd_queue[cmd_queue_head].private;
	    if ( revese_cnt > 5){
            uint8_t *address = (cmd_queue[cmd_queue_head].private);
            if (address != NULL) v = address[i];
            else v = SENDPAGE_ERROR_NO_PAGEONID;
	    } else if ( revese_cnt == 3 ) v = SEND_PAGE_STUFFED_TYPE;//triggerlog type
	    else if ( revese_cnt == 4 ) v = (uint8_t)addr_or_page ;
	    else v = addr_or_page >> 8;

	  ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
	} else { /* i == size-3 or size-2*/
		v = (i == (size-2) ) ? com.commcrc & 0xff : com.commcrc >> 8;/* CRC sent in little endian as in SIPR*/
		ringbuf_push_bytestuff(&comm_sndbuf, v); /* When sending crc itself => doNT update crc*/
	}

    return 'P';
}


/* sends RT data in triggerlog format
*  MT_RTVAR_COUNTMORE byte data+ 1 bytetriggerlog type(32) + 2 byte crc + 1 byte HDLC_END
*/
uint8_t MTsendRTvar_triggerlog(uint8_t i)
{
    uint8_t v;
    if( i <= MT_RTVAR_COUNTMORE ){
        if (i == MT_RTVAR_COUNTMORE) v = RT_STUFFED_TYPE;
        else{
            v = MTsendRTvar(i, 0);
	}
	ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    }else{
        if( i >= (MT_RTVAR_COUNTMORE + 3) ){ // MT_RTVAR_COUNTMORE +3
		ringbuf_push(&comm_sndbuf, HDLC_END); // send framing char, do NOT stuff
		com.commcrc = 0xffff;
        } else { // MT_RTVAR_COUNTMORE +1 or +2
		v = (i == (MT_RTVAR_COUNTMORE + 1) ) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
    }
    return 0;
}

#define ATOMIC_HELPER (uint8_t *)(caller==0 ? &cmd_queue[cmd_queue_head].privch.a : &(com.atomic_helper))
/**
* i: index 0..len-1
* if caller==0 store ATOMIC_HELPER in cmdqueue, otherwise in global var
*/
uint8_t PROGMEM COMM_AD_CHANNELS[8]={4,3,6,7,5,0,1,2};//send and ad order does not match

/// i must be monotonous increasing
uint8_t MTsendRTvar(uint8_t i, uint8_t caller) {
//helper for analog value send in the historic order, todo send in real order
if ( (i>=24) && (i< 40)){ // 24 to 39
        uint8_t channel;
#ifdef KNOCK_ALTERNATIVE
        if (alternative_rtdata() && (i >= 32) ) return knock_cyl[i-32].knock_adjust; // knock_cyl[0] to [7]
#endif

        i -= 24;
        channel = pgm_read_byte( &COMM_AD_CHANNELS[i >> 1]);//get the historic order
        if(i & 0x01){
            return *ATOMIC_HELPER;
        } else { // even index first, store the following second(odd) byte
		*ATOMIC_HELPER = getMCPdataLOW(channel);
            return getMCPdataHIGH(channel);
    }
    }

  switch (i) {
  case 0:
    return (uint8_t)rtc.sec;  //secl

  case 1:
/* TODO:
  check this.
  "squirt" is apparently a bitmap of the something:
  ....BBAAba
  where
  capital A, B : ????
  lowercase a, b: if the given channels are actually squirting
  // each squirt_t has its own status field, merge these fields
    a = inj_port1.status;
    b = inj_port2.status;

    return ((a & 0x01)
    		| ((a & 0x06) << 1)
           	| ((b & 0x01) << 1)
           	| ((b & 0x06) << 3));
*/
	return engine.boostalt_command_duty;
  case 2:
// remove the 'idle' bit to avoid confusing mstweak: return (engine.status & (uint8_t)~_BV(idle));  //engine
	  return engine.status;  //engine

  case 3: return engine.kpa_or_tps;  // AlphaN load; originally sensors8(BARO)
  // high resolution MAP
  case 4: {
      *ATOMIC_HELPER = engine.kpa250 & 0xFF;
      return engine.kpa250 >> 8;
  }
  case 5: return *ATOMIC_HELPER;

  case 6: return engine.mat;   //mat
  case 7: return engine.coolant;   //clt
  case 8: return engine.tps;   //tps
  case 9: return engine.batt;  //batt
  case 10: // o2
#ifdef WBO2
	if (WBO2_SENSOR_CONFIGURED)
	{
// traditional stuff via lookup table
//        	return wbo2.Ip_output;
		/* make a concept. Make sure gauge displays 20.9 for 20.9% O2
		* eg. gauge means lambda 0.6 .. 5 than O2% between 5..25
		* Refer to megatune/vemstune compatible lookup table file
		* Test it. DO NOT NEGLECT THESE INSTRUCTIONS.
		* Anyone who changes this without the above, most significantly without:
		* written concept, working and tested tablefile BOTH for vemstune and megatune
		* or if it does not show 20.9 for 20.9% O2, will be prevented from accessing SVN
		*/

		// Andrey's solution, precise values required for AutoTune process possible, and perfect for dataloging
		// 20.9 already showed with StallO2 (see PW bytes 16:17), therefore 20.9 not needed to seen here
		// MT calcs lambda with simple scale/offset without any lookup tables
		// MT vems.ini uses: lambda = {(egoADC + 306)/470}
		// So rt_byte_10=164 means lambda = 1.0
		// ((lambda * 14.7 * 1024) / 32 ) - 306
		// 0.7 : 23.28
		// 1.0 : 164.4
		// 1.1 : 211.44
		// 1.19 : 253.77
        // 1.25 : 282 (255)
		// new implementation changed above 1.1 lambda:
		// 1.1 : 211.22 ??
		// 1.4 : 281.78 (255)
		// 1.28 : 253.56
		uint16_t t16;
		t16 = wbo2.current_afr>>5;  // 15053 -> 471   (lambda 1.1 = 517)
        if( t16 <= 306 ) return 0; // -306 below would underflow
        if( t16 <= 517 ) return t16-306; // lambda range is 0.65...1.1 (1..211)
        t16 -= 95;
        t16 >>= 1;
        return t16 < 256 ? t16 : 255; // above 1.1 .. 1.28
	}
	else
#endif
		return ego_voltage();  //ego
  case 11: return mult16_8(corr.ego16,100)>>8;	//egocorr
  case 12: return corr.air;       //aircorr
  case 13: return (corr.warm >=  1535) ? 255 : (mult16_8(corr.warm-230,50)>>8);      //warmcorr railed to 600%

  // high resolution RPM
  case 14: {
      *ATOMIC_HELPER = engine.rpm_vhr & 0xFF;
      return engine.rpm_vhr >> 8;
  }
  case 15: return *ATOMIC_HELPER;

//show O2 instead pw in meatune for mde02 sensor calibration with stalled engine
  case 16: {
    *ATOMIC_HELPER = (engine.calculated_fuelpw & 0xFF);
    return ((engine.rpm==0) && (debug.enable_flags & _BV(1)) && wbo2.current_o2) ? wbo2.current_o2 >> 8 : (engine.calculated_fuelpw >> 8);     //pw high
  }
  case 17: return ((engine.rpm==0) && (debug.enable_flags & _BV(1)) && wbo2.current_o2) ? wbo2.current_o2 & 0xFF: *ATOMIC_HELPER;   //pw low
  case 18: return corr.baro;      // barocorr is 128 based
  case 19: {	// calculate from the stored high-res value (/100 and railing is needed)
		uint16_t ge;
		ge = mult16_8(corr.gammae381, 68) >> 8;// 381 * 68 / 256 = 101 (NOT 100 !)

		if ((ge >> 8) != 0)  // check for overflow
			return 0xFF;
		else
			return (uint8_t)ge;
	}
  case 20: return corr.ve_hp;

// should we use this style ?
// #define MT_RT_OFFSET_XXX (21)
// #define MT_RT_OFFSET_IGNADV (MT_RT_OFFSET_XXX+1)

  case 21: return engine.dwell>>4;	// 64 usec resolution (possibly but unlikely overflowing)
  case 22: return (engine.ignadv < 510 ? (engine.ignadv>>1) : 255); // 128 + ignadv_halfdeg

  case 23: return iac.position;

// make sure 16 bit variables start on even offsets:

// raw egt data, do scaling externally (in megatune based on egtofs and egtcal)

  case 40:
		  return corr.lambda;
#ifdef SECOND_VSS
  case 42: return wheel_speed2.wspeed;
#endif // SECOND_VSS

#ifdef A340E //andrey always use alt_knock
  case 41: return at_control.at_temp;
#else
  // dump knock related variables for cylinder 0
  case 41: return knock_cyl[0].knock_val >> 8;
#endif
#ifdef KNOCK_ALTERNATIVE
  case 43: return knock_cyl[0].mean_val >> 8;
  case 44: return knock_cyl[0].mean_val & 0xFF;
#else
  case 43: return knock_cyl[0].noise_val >> 8;
  case 44: return knock_cyl[0].noise_val & 0xFF;
#endif
  case 45: return knock_cyl[0].knock_adjust;
  case 46: return knock_cyl[0].state;

  // status bytes like WBO2 activated and alike
  case 47: {// bit1 unused later used for trigger error, no it send at byte 66
	uint8_t status = 0;
#ifdef ALS
	if (engine.status_ext & _BV( als_switch_activated )) status |= 0x80;	// als switch active
	// if (is_als_active()) status |= 0x01; // als function active
#endif
	if( engine.status_trig & _BV(n2o_active)) status |= 0x01; // N2O active
#ifdef SOFTLAUNCH
	if ( is_launch_input_or_function_active() ) status |= 0x40; // launch control active or launch input button active
#endif
	if ( during_actual_shiftcut() ) status |= 0x20; // shift cut
	if (engine.status_trig & _BV( injector_staging_active )) status |= 0x10; // inj stage2 active
//	if (engine.status & _BV(idle)) status |= 0x08;	// idle control
	if (engine.status4 &_BV(STATUS4_CONFSW) ) status |= 0x08;	// Config slot
	if (engine.status_ext & _BV(o2_closed_loop)) status |= 0x04; // closed loop (ego)
	if (eeprom.flags & _BV(active)) status |= 0x02; // eeprom saving in progress flag. Since 1.1.83
	return status;
  }
#ifdef BOOSTCONTROL
	// without boostcontrol the default 0 will be returned
	case 48: return boost_data.boost_target16 >> 5; // 1/8 kPa => 4 kPa
	case 49: return engine.boost_command_duty; // boost solenoid duty
	case 50: return (uint8_t)((boostcontrol_pid.integral+0x4000)>>7); // boost PID integral, assume PID_INTEGRAL_EXP=6
#endif
	case 51: return (uint8_t)(iac.integral>>8)+0x80; // iac.integral, assume INTEGRAL_LIMIT_SCALE=7
	case 52: engine.status_ext &= ~_BV(individual_allowed); return engine.sparkcutlvl; // 0 .. 16 ( 0 is no igncut, normal operation)

#ifndef KNOCK_ALTERNATIVE
	case 53: return individual_power[0];
	case 54: return individual_power[1];
	case 55: return individual_power[2];
	case 56: return individual_power[3];
	case 57: return individual_power[4];
	case 58: return individual_power[5];
	case 59: return individual_power[6];
	case 60: engine.status_ext |= _BV(individual_allowed);
		if (debug.enable_flags & _BV(7)){
			pushfunc((void*)MTsendRTvar, MT_RTVAR_COUNT);
		} //TODO:
		return (corr.tpsaccel >= 2040) ? 0xFF: (corr.tpsaccel>>3);  //tpsaccel, 4usec => 32 usec ; set back to individual_power[7];
#else
	case 53:
		if (alternative_rtdata()) return knock_cyl[0].knock_val&0xFC00 ? 0xFF : knock_cyl[0].knock_val >> 2;
		else return individual_power[0];
	case 54:
		if (alternative_rtdata()) return knock_cyl[1].knock_val&0xFC00 ? 0xFF : knock_cyl[1].knock_val >> 2;
		else return individual_power[1];
	case 55:
		if (alternative_rtdata()) return knock_cyl[2].knock_val&0xFC00 ? 0xFF : knock_cyl[2].knock_val >> 2;
		else return individual_power[2];
	case 56:
		if (alternative_rtdata()) return knock_cyl[3].knock_val&0xFC00 ? 0xFF : knock_cyl[3].knock_val >> 2;
		else return individual_power[3];
	case 57:
		if (alternative_rtdata()) return knock_cyl[4].knock_val&0xFC00 ? 0xFF : knock_cyl[4].knock_val >> 2;
		else return individual_power[4];
	case 58:
		if (alternative_rtdata()) return knock_cyl[5].knock_val&0xFC00 ? 0xFF : knock_cyl[5].knock_val >> 2;
		else return individual_power[5];
	case 59:
		if (alternative_rtdata()) return knock_cyl[6].knock_val&0xFC00 ? 0xFF : knock_cyl[6].knock_val >> 2;
		else return individual_power[6];
	case 60: engine.status_ext |= _BV(individual_allowed);
		if (debug.enable_flags & _BV(7)){
			pushfunc((void*)MTsendRTvar, MT_RTVAR_COUNT);
		} //TODO:
		if (alternative_rtdata()) return knock_cyl[7].knock_val&0xFC00 ? 0xFF : knock_cyl[7].knock_val >> 2;
		else return (corr.tpsaccel >= 2040) ? 0xFF: (corr.tpsaccel>>3);  //tpsaccel, 4usec => 32 usec ; set back to individual_power[7];
#endif
	case 61: return engine.tok_diff; // this dV/dt will stay
#ifndef  PS2KEYBOARD
	case 63:
		{
		uint8_t t8=0;
#ifdef A340E
		if (at_control.at_status&_BV(LOCKUP))
			t8=100;
#endif
		return engine.gear+t8;	//gear
		}
	case 62: return wheel_speed1.wspeed;  //speed
#else
	case 62: return now_64msec() - engine.launchinact_time;  //elapsed time
#endif
#ifdef CAMSHAFT_ANGLE
    case 64: return engine.sectrig_angleI;
    case 65: return engine.camshaft_dcI >> 4;
#endif
    case 66: //status2 (trigger error bits (bit 4:6). Lower half and bit7 unused
        {
        uint8_t status=0;
        if (engine.trigger_error & 0x0f){//lower 4 bit need to turn on error for 1.6sec if zero error code invalid
            status = (engine.trigger_error & 0x70 ); //bit 7 reserved for future trigger error code
        }
        return status;
        }
#ifdef CAMSHAFT_ANGLE_EXHAUST
    case 67: return engine.extrig_angle;//third trigger angle
    case 68: return engine.camshaft_dcE >> 4; // camsahft exhaust control duty cycle
#endif
	default: return 0;
  }
}

/** send 32 bytes:
 * first 16 bytes is the 8 mcp3208 in big endian
 * 2nd 16 bytes is the 7 internal AD in big endian, but channel 7 first (0 latest)
 */
uint8_t allSensorRaw(uint8_t i)
{
	uint8_t t8, *p;
	if( i < 16 ){ // 0 .. 15
		t8 = i >> 1;
		if( i & 1 ){ // LOW sent AFTER high byte
			return getMCPdataLOW( t8 );
		} else { // send high FIRST:
			return getMCPdataHIGH( t8 );
		}
	}
	p = (uint8_t*)sensor_filt; // convert to byte array
	return p[ 31 - i ];
}

/* Dump the table
* private higher byte holds the table identifier char,
* private lower byte holds the table-line
* line higher 4 bits contain the counter (eg 0..7)
* line lower 4 bits contain the maxvalue of counter (eg 7)
* count (conservatively) holds the character inside the line
*/
uint8_t TableDump(uint8_t cnt){
	uint16_t t;
	char table;
	uint8_t table_lines; // number of table lines without the r[] k[] lines.
	uint8_t line; // 0..15, counts up after every line. if(line > table_lines-1) than .[r]=.. or .[k]=.. printed
	uint8_t linemax; // 0..15, stays same determined by menu.c TableDump "caller"/initializer
	uint8_t *p;
	uint8_t len; // elements/line. Higher than 16 is allowed (if line length is less than 240 characters)
	uint8_t line_end; // calculated: HEADER_LENGTH + len * block_printlen; Must be less than 240, ~len<78

// if 2 is allowed, than allow case 3,4 in the last switch-case
#define block_bytes 1
#define block_printlen 3

	t = (uint16_t)(cmd_queue[cmd_queue_head].private);
	table = t>>8;

	if(cnt >= 0xF0){
		// similar to state 0, but 0 cannot be used again for some reason...
		cmd_queue[cmd_queue_head].count = 0;
		return table;
	}

	line = t & 0xFF;
	linemax = line & 0x0F;
	line >>= 4;
	// block_printlen = 3;
	// block_bytes = 1;

	table_lines = 0xFF; // this way line > table_lines condition not triggered

	switch(table){ // this should get from a place common with menu
		case 'j':
			p=ve.ve8_table;
			len = VE_SIZE_RPM;
			table_lines = VE_SIZE_KPA;
			break;
		// case 'k':
		//	p=ve.kpa_range;
		//	len = VE_SIZE_KPA;
		//	break;
		case 'l':
			p=ve.lambda;
			len = LAMBDA_SIZE_RPM;
			table_lines = LAMBDA_SIZE_KPA;
			break;
		// either ifndef NOIGN is not needed, or needed in menu.c too
		// I allow the read-write of 'n' table for other applications...
		case 'n':
			p=ign.adv_table;
			len = IGNADV_SIZE_RPM;
			table_lines = IGNADV_SIZE_KPA;
			break;
#ifdef HWMAP
		case 'h':
			p=hwmap.table;
			len = HWMAP_SIZE;
			break;
#endif

#ifdef A340E
		case 'u':
			p = at_config.shift_up[0];
			len = sizeof(at_config);
			break;
#endif

#ifdef TOYOTA_1JZ_VVTI
		case 'z':
			p = vvti_map.vvti;
			len = VE_SIZE_RPM;
			break;
#endif

#ifdef KNOCK_ALTERNATIVE
		case 'f':
			p = knock_table.knock_noise;
			len = KNOCK_SIZE_RPM;
			break;
#endif

#ifdef BOOSTCONTROL
		case 'b':
			p = boost.boostmaxtarget_table;
			len = BOOST_SIZE_RPM;
			table_lines = BOOST_BTABLE_GEARLINES;
			break;
		case 't':
			p = boost.tpstransfer_table;
			len = TPSTRANSFERTABLE_SIZE;
			break;
		case 'y':
			p = boost.boost_ref_table;
			len = IGNADV_SIZE_RPM;
			table_lines = BOOST_SIZE_KPA;
			break;
#endif
#ifdef MAT_ENRICH
		case 'e':
			p = ve.mat_enrich;
			len = 8;
			table_lines = LAMBDA_SIZE_KPA;
			break;
#endif
#ifdef CAMSHAFT_ANGLE
#ifdef TOYOTA_1JZ_VVTI
#error "CAMSHAFT_ANGLE incompatible with TOYOTA_1JZ_VVTI"
#endif
		case 'z':
			p = ve.camshaft_angleI;
			len = 16; // theoretically 16+2 lines with 8 in each, but hack-sent this way
			table_lines = 8;
			break;
#endif
#ifdef CHARGE_TEMP
		case 'e':
			p = f_charge.f_charge;
			len = VE_SIZE_RPM;
			table_lines = VE_SIZE_KPA;
			break;
#endif
#ifdef MAP_CORR
		case 'i':
			p = ve.map_corr;
			len = MAP_CORR_SIZE_RPM;
			table_lines = MAP_CORR_SIZE_KPA;
			break;
#endif
		case 's':
			p = ve.launch_rpmtable;
			len = 8;
			table_lines = 2;
			break;
		case 'p':
			p = boost.anytrim1_array;
			len = 17; // 3*17 bytes
			table_lines = 3;
			break;
		case 'w':
		default:
			p=ve.mat_retard;
			len = LAMBDA_SIZE_RPM;
			table_lines = LAMBDA_SIZE_KPA;
			break;
	}

#define HEADER_LENGTH (5)
	if(cnt < HEADER_LENGTH){ // line beginning
		switch(cnt){
			case 0: return table;
			case 1: return '[';
			case 2: if( line < table_lines ){ // print kpa line as j[k]=xx xx xx ...
					return hex_lo(line);
				} else if( line == table_lines ){
					return 'r';
				} else return 'k';
			case 3: return ']';
			case 4: return '=';
		}	// switch
	}	// if

	// we must determine how long the line is for the given table:
	// 3 * len or 5 *len for a 16 bit table
	// (-1 in any case, since no space before first, see the '\r' branch)
	if( table == 'j' && VE_SIZE_KPA == 14 && line == 13+2 ){
		// for 16x14 j table last (kpa) line is 14 bytes only
		line_end = 14;
	} else line_end = len;
	line_end *= block_printlen;
	line_end += HEADER_LENGTH;

	if(cnt == line_end){ // shell we finish or go to next line
		if(line != linemax){ // not finished
			// endianess dependent and bigger flash: uint8_t *little_endian_p; little_endian_p = &cmd_queue[cmd_queue_head].private; * little_endian_p += 16;
			// line++;
			// solution that does not assume little endian:
			cmd_queue[cmd_queue_head].private += 16; // = (void*)((table<<8) + (line<<4) + linemax);
			cmd_queue[cmd_queue_head].count = 0xF0-1; // will be incremented (0-1 doesnt work)
		} else {
			if( table == 'e'){//this condition + call adds +178 byte to flash !?!
				// 0xF240 - 0xF18E = 178 = 0xF6C4 - 0xF612 (with either my_make)
				// moving pushfuncTableDump() above did not help either
				pushfuncTableDump( TABLES_TO_DUMP2, sizeof(TABLES_TO_DUMP2) >> 1 );
			}
			cmd_queue[cmd_queue_head].count = 0xFE; // set to len => THE END
		}
		return '\n';
	}

	if(cnt == line_end - 1){
		return '\r';
	}

	cnt -= (HEADER_LENGTH - 1);

	p += len * line * block_bytes;

	// calculate modulus without division:
	while(cnt >= block_printlen){
		cnt -= block_printlen;
		p += block_bytes;
	}

	switch(cnt){	// every byte is printed to 3 characters
		// case 0: return ' '; // handled by default branch
		case 1: return hex_hi(*p);
		case 2: return hex_lo(*p);
#if(block_bytes>1)
		case 3: return hex_hi(*(p-1));	// only for 16 bit table (little endian)
		case 4: return hex_lo(*(p-1));
#endif
		default: return ' ';
	}
}

/* Dump the config structure:
 * We keep track of line number in private.
 * position inside the line is in count, but somewhat tricky
 * Note that len is 0xFE because of pushfunc() hack, although we specified 0xFF
 */
uint8_t ConfigDump(uint8_t cnt) {
	uint8_t c, ch, i, *str;
	uint8_t last_element;
	uint16_t idx;
	PGM_VOID_P P;
	// not always needed, but almost:
	// if(cnt == 0xF7) cmd_queue[cmd_queue_head].count = cnt = 0; // needed because of the len-1 hack

	idx = (uint16_t)(cmd_queue[cmd_queue_head].private);
	ch = cmd_queue[cmd_queue_head].privch.a;
	if ( ch == 'C') {
	    P = (PGM_VOID_P)(varstr_ext + idx);
	    last_element = (sizeof(struct config_ext_t)-1);
	}
	else if( ch == 's') {
	    P = (PGM_VOID_P)(varstr_shared + idx);
	    last_element = (sizeof(struct config_shared_t)-1);
	}else{
	    P = (PGM_VOID_P)(varstr + idx);
	    last_element = (sizeof(struct config_t)-1);
	}
	i = 0;
	while((str=(void*)PRG_RDW(P)) == NULL){
		P -= 2;  // addresses are two-byte
		i++;
	}

// #define LAST_ELEMENT (sizeof(struct config_t)-1)

	if(cnt < 0xF8) {
		c = PRG_RDB(str + cnt);
		if(c) {
			return c;
		} else { // end of string
		// we need the [.]= if i>0 or the next pointer is NULL but not at the very end
			if(i || (((void*)PRG_RDW(P+2)) == NULL && idx < last_element)){ // no need for [i]=.. just =..
				cmd_queue[cmd_queue_head].count = 0xF8 - 1 ; // will be incremented!
				return '[';
			} else {
				cmd_queue[cmd_queue_head].count = 0xFB - 1 ; // will be incremented!
				return '=';
			}
		}
	}

    if (ch == 'C') str = ((uint8_t *)(&config_ext));
    else if (ch == 's') str = ((uint8_t *)(&config_s));
	else str = ((uint8_t *)(&config));

	switch (cnt){
		case 0xF8: return hex_ch(i);
		case 0xF9: return ']';
		case 0xFA: return '=';
		case 0xFB: return hex_hi(*((uint8_t *)str + idx));
		case 0xFC: return hex_lo(*((uint8_t *)str + idx));
		case 0xFD: return '\r';
		case 0xFE: // let's see: are we at the very end ?
			if((uint16_t)(++ cmd_queue[cmd_queue_head].private) <= last_element ){
				// cmd_queue[cmd_queue_head].count = 0xFE;	// end of the show
				// = cmd_queue[cmd_queue_head].len
				cmd_queue[cmd_queue_head].count = 0xFF; // show must go on next line
			};
			return '\n';
	}
	return 'z';
}

/***************************************************************************/
/***************************************************************************/
/*** End of MegaTune compatibility functions                             ***/
/***************************************************************************/
/***************************************************************************/




/***************************************************************************/
/*** sendConfigVar                                                       ***/
/***************************************************************************/
uint8_t sendConfigVar(uint8_t i) {
  if (i < sizeof(struct config_t))
    return *( (uint8_t *)&config + i );
  else
    return 0;
}


/*
 * returns upper and lower byte of "private"
 */
uint8_t srl_direct( uint8_t i ){
	uint16_t t = (uint16_t) cmd_queue[cmd_queue_head].private;
	switch(i){
		case 0: {
			return t>>8;
		}
		case 1: {
			return t & 0xff;
		}
		default: return 'z';
	}
}

/** for .private=0xCDAB
 we print ABCD if len=4
 or AB if len=2
*/
uint8_t srl_direct_hex( uint8_t i ){
	uint16_t t = (uint16_t) cmd_queue[cmd_queue_head].private;
	switch(i){
		case 0: return hex_hi(t & 0xFF);
		case 1: return hex_lo(t & 0xFF);
        case 2: return hex_hi(t >> 8);
		case 3: return hex_lo(t >> 8);
		default: return 'z';
	}
}

/************************************
* generic serial logging function.
* uses private from the actual command as a pointer (array base).
* Unfortunately we need to know the string-length in advance
* For constant(-length) strings this is OK.
* We could set a (global or static variable is simplest) flag
* to notify the caller about the end of transmission.
* This way other comm functions could be left untouched.
************************************/
uint8_t srl_generic(uint8_t addr)
{
	return (cmd_queue[cmd_queue_head].private)[addr];
}

uint8_t srl_generic_atomic16_crc(uint8_t addr)
{
    static uint8_t upper_byte;
    uint8_t t8;
    if (addr == 0) com.commcrc = 0xffff;

    if (addr >= DBG_LEN ){
        if (addr == DBG_LEN ) return com.commcrc & 0xff;//low byte
        return com.commcrc >> 8;//high byte
    }

    if (addr & 0x01){//odd memory address
        com.commcrc = _crc16_update( com.commcrc, upper_byte);
        return upper_byte;
    }//else even address
    upper_byte = (cmd_queue[cmd_queue_head].private)[(addr + 1)];//save next byte to atomic read
	t8 = (cmd_queue[cmd_queue_head].private)[addr];
    com.commcrc = _crc16_update( com.commcrc, t8);
	return t8;
}

/* same as above, but private points to PROGMEM (flash), eg. PSTR("...") */
uint8_t srl_generic_P(uint8_t addr)
{
	return PRG_RDB(&(cmd_queue[cmd_queue_head].private[addr]));
}

/* sends data direct from EEPROM in triggerlog format
*  240 byte data+ 1 bytetriggerlog type(16) + 2 byte crc + 1 byte HDLC_END
*  if EEPROM busy returns with CRC fail
*/
uint8_t srl_generic_E(uint8_t i)
{
    uint8_t v;
    if( i < (TRIGGERFRAME_READ_EEPROM_FIX + 2 ) ){
        if (i == ( TRIGGERFRAME_READ_EEPROM_FIX + 1 ) ) v = READ_EEPROM_STUFFED_TYPE_COMMAND-128;//answer type
        else if ( i == TRIGGERFRAME_READ_EEPROM_FIX ) v = (uint8_t)(uint16_t)cmd_queue[cmd_queue_head].private;
        else{
		if (eeprom.flags & _BV(active)) v = 0;//EEPROM busy send zeros
		else v = eeprom_read_byte((uint8_t*)((uint16_t)(cmd_queue[cmd_queue_head].private)*16+i));
	}
	ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    }else{
        if( i >= ( TRIGGERFRAME_READ_EEPROM_FIX + 4 ) ){
		ringbuf_push(&comm_sndbuf, HDLC_END); // send framing char, do NOT stuff
		com.commcrc = 0xffff;
        }else { // 3 or 4
		if (eeprom.flags & _BV(active)) v = 0;//EEPROM busy send bad crc
		else v = (i == ( TRIGGERFRAME_READ_EEPROM_FIX + 2 ) ) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
    }
    return 0;
}

// helper funtion used to access flash above the 16 bit address limit
// TODO: check that it works
uint8_t srl_generic_P_upper(uint8_t addr)
{
	uint32_t v;
	v = 0x10000 | (uint16_t)&cmd_queue[cmd_queue_head].private[addr];
	return pgm_read_byte_far(v);
}

void change_page(uint8_t page) {
	if (page > MT_PAGE_IDX_LAST) {
		page = MT_PAGE_IDX_DUMMY;
	}
	debug.megatune_page = page;
}

/** ===============================================
 * make sure comm.h MT_PAGE_IDX_... and pages_size match.
 * Changing these breaks some tuningsoftwares (SIPR ??)
 */

#define MT_PAGE_PTR_DUMMY (uint8_t *)NULL

PROGMEM uint8_t *pages_ptr[] = {
 	(uint8_t *)&config
#ifdef HWMAP
 	,(uint8_t *)&hwmap.table
#else	//  sizeof(hwmap) bytes _must_ be transmitted to megatune
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef BOOSTCONTROL
	,(uint8_t *)&boost.tpstransfer_table
#else // no BOOSTCONTROL
 	, MT_PAGE_PTR_DUMMY
#endif
 	,(uint8_t *)&config_ext
#ifdef MAT_ENRICH
	,(uint8_t *)&ve.mat_enrich
#else
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef A340E
 	,(uint8_t *)&at_config
#else
 	, MT_PAGE_PTR_DUMMY
#endif
	, ve.launch_rpmtable
    ,(uint8_t *)&config_s
    ,(uint8_t *)&ve.lambda
	,(uint8_t *)&ve.lambda // lambdaview !!!
 	,(uint8_t *)&ve.ve8_table
#ifdef CHARGE_TEMP
 	,(uint8_t *)&f_charge.f_charge
#else
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef MAP_CORR
	,(uint8_t *)&ve.map_corr
#else
 	, MT_PAGE_PTR_DUMMY
#endif
	,(uint8_t *)&ve.mat_retard
	// following tables all share rpmkpa with ign_adv table:
	,(uint8_t *)&ign.adv_table
#ifdef BOOSTCONTROL
	,(uint8_t *)&boost.boost_ref_table
#else
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef TOYOTA_1JZ_VVTI
 	,(uint8_t *)&vvti_map.vvti
#else
 	, MT_PAGE_PTR_DUMMY
#endif
#ifdef KNOCK_ALTERNATIVE
	,(uint8_t *)&knock_table.knock_noise
#else
 	, MT_PAGE_PTR_DUMMY
#endif
};

PROGMEM uint8_t pages_size[] = {
 	sizeof(struct config_t)
 	, HWMAP_BYTES
#ifdef BOOSTCONTROL
	,sizeof(boost.tpstransfer_table) + sizeof(boost.boostmaxtarget_table)
	+ sizeof(boost.boostpidrefpwm_table) + sizeof(boost.boostgear_rpmrange)
	+ sizeof(boost.anytrim1_array) + sizeof(boost.rotary_trailing_spark )
	+ sizeof(boost.inj_angle)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
	,sizeof(struct config_ext_t)
#ifdef MAT_ENRICH
	,224 //80 mat table + 2*64cam_table+16cam headers NOT sizeof(ve.mat_enrich) because PAGE_IDX_FIRST_RPMKPA
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef A340E
 	,sizeof(at_config)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
	,sizeof(ve.launch_rpmtable)
    ,sizeof(config_s)
	,sizeof(ve.lambda)
 	,sizeof(ve.lambda)
 	,sizeof(ve.ve8_table)
#ifdef CHARGE_TEMP
 	,sizeof(f_charge.f_charge)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef MAP_CORR
	,sizeof(ve.map_corr)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
	,sizeof(ve.mat_retard)
	// following tables all share rpmkpa with ign_adv table:
	,sizeof(ign.adv_table)
#ifdef BOOSTCONTROL
 	,sizeof(boost.boost_ref_table)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef TOYOTA_1JZ_VVTI
 	,sizeof(vvti_map.vvti)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
#ifdef KNOCK_ALTERNATIVE
 	,sizeof(knock_table.knock_noise)
#else
	, MT_PAGE_SIZE_DUMMY
#endif
};

uint8_t MEGATUNE_PAGE_SIZE(uint8_t page)
{
    return pgm_read_byte(pages_size + page);
}

uint8_t* MEGATUNE_PAGE_POINTER(uint8_t page)
{
    return (uint8_t*)pgm_read_word(pages_ptr + page);
}

/*
* calculates the 16 bit crc for megatune page with RPM/KPA header
*/
uint16_t megatune_page_crc(uint8_t idx)
{
    if (  idx > MT_PAGE_IDX_LAST) return 0xffff;//id not valid
    uint8_t cnt;
    uint8_t* page_ptr = MEGATUNE_PAGE_POINTER(idx);
    uint8_t page_len = MEGATUNE_PAGE_SIZE(idx);
    uint16_t crc = 0xffff;

    if ( page_ptr == MT_PAGE_PTR_DUMMY ) return 0xffff;
    for ( cnt=0 ; cnt<page_len ; cnt++){
        crc = _crc16_update( crc, page_ptr[cnt]);
    }
    if (!(page_has_rpmkpa(idx))) return crc;

    struct bins_pointer_t bins_p;
	get_bins_pointer(idx, &bins_p);

    page_ptr = bins_p.p_rpm_range;
    page_len = bins_p.rpm_size;
    for ( cnt=0 ; cnt<page_len ; cnt++){
        crc = _crc16_update( crc, page_ptr[cnt]);
    }
    page_ptr = bins_p.p_kpa_range;
    page_len = bins_p.kpa_size;
    for ( cnt=0 ; cnt<page_len ; cnt++){
        crc = _crc16_update( crc, page_ptr[cnt]);
    }
    return crc;
}

/*
* returns the length for megatune page with RPM/KPA header
*/
uint8_t megatune_page_len(uint8_t idx)
{
    if (  idx > MT_PAGE_IDX_LAST) return 0;//id not valid
    uint8_t page_len = MEGATUNE_PAGE_SIZE(idx);
    if ( page_len == MT_PAGE_SIZE_DUMMY ) return 0;
    if (!(page_has_rpmkpa(idx))) return page_len;

    struct bins_pointer_t bins_p;
	get_bins_pointer(idx, &bins_p);
    page_len += bins_p.rpm_size;
    page_len += bins_p.kpa_size;
    return page_len;
}

/*
* sends the data from the private function with byte stuffing
* works only with parameterless private function which don use crc
*/
uint8_t sendStuffedSimple(uint8_t cnt)
{
    uint8_t size= cmd_queue[cmd_queue_head].len;
    uint8_t v;
    if (cnt <= (size-4)){
        if (cnt == (size-4)){
            v= SIMPLE_MSG_STUFFED_TYPE;
        }else{
            v= (cmd_queue[cmd_queue_head].privatefunc)(cnt);
        }
	  ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    } else if(cnt >= size-1){
		ringbuf_push(&comm_sndbuf, HDLC_END); // send framing char, do NOT stuff
		com.commcrc = 0xffff;
	} else { // i == 3 or 4
		v = (cnt == (size-3)) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian as in SIPR
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
	}

	return 's';
}

/*
* sends all page len and crc in triggerframe format
*/
uint8_t send_page_len_crc_stuffed (uint8_t cnt)
{
    uint8_t v;
    if( cnt <= MT_PAGE_CRC_PACKET_LEN-4 ){
        if (cnt == MT_PAGE_CRC_PACKET_LEN-4) v = PageCrc_stuffed_type;
        else if ( cnt < (2*(MT_PAGE_IDX_LAST+1))){//send all page crc
            if ( (cnt & 0x01) == 0){//even means calulate crc, store high byte and send low byte
                uint16_t crc = megatune_page_crc(cnt>>1);
                cmd_queue[cmd_queue_head].privch.a = crc >>8;
                v = (uint8_t)crc;
            }else{// just send the prestored crc low byte
                v = cmd_queue[cmd_queue_head].privch.a;
            }
        }else{ //send all page len
            v = megatune_page_len(cnt-(2*(MT_PAGE_IDX_LAST+1)));
	}
	ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
    }else{
        if( cnt >= (MT_PAGE_CRC_PACKET_LEN-1) ){
		ringbuf_push(&comm_sndbuf, HDLC_END); // send framing char, do NOT stuff
		com.commcrc = 0xffff;
        } else { // MT_PAGE_CRC_PACKET_LEN-1,-2
		v = (cnt == (MT_PAGE_CRC_PACKET_LEN-3) ) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
    }
    return 'C';
}

/***************************************************************************/
/*** comm                                                                ***/
/***************************************************************************/
void comm(uint8_t rx_data) {
  static uint8_t rx_offset = 0;
  static uint8_t rx_cmd = 0;

  switch (com.comm_state) {
  case OFFSET:
    rx_offset = rx_data;  // the offset for storing the new byte
    com.comm_state = VALUE;
    break;

  case VALUE:
    switch (rx_cmd) {
	  /* Separate command for each page write / read
	  * => No config/other overwrite on ecu reset. */
    case 'X': // forcing: channel + value
		debug.forced_channel = rx_offset;
		debug.forced_value = rx_data;
		break;
	  case 'O':
		change_page(MT_PAGE_IDX_CONFIG);
		store_page(rx_offset, rx_data);
		break;
	  case 'G':
		change_page(MT_PAGE_IDX_CONFIGEXT);
		store_page(rx_offset, rx_data);
		break;
	  case 131:
		change_page(MT_PAGE_IDX_CONFIG_SHARED);
		store_page(rx_offset, rx_data);
		break;
	  case 'H':
		change_page(MT_PAGE_IDX_HWMAP);
		store_page(rx_offset, rx_data);
		break;
	  case 'T':
	    change_page(MT_PAGE_IDX_BOOST);
		store_page(rx_offset, rx_data);
		break;
      case 'J':
		change_page(MT_PAGE_IDX_VE);
		store_page(rx_offset, rx_data);
		break;
      case 'N':
		change_page(MT_PAGE_IDX_IGNADV);
		store_page(rx_offset, rx_data);
		break;
      case 'L':
		change_page(MT_PAGE_IDX_LAMBDAVIEW);
		store_page(rx_offset, rx_data);
		break;
	case ')':
		change_page(MT_PAGE_IDX_LAMBDA);
		store_page(rx_offset, rx_data);
		break;
#ifdef A340E
	  case 'U':
		change_page(MT_PAGE_IDX_A340E);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef TOYOTA_1JZ_VVTI
	  case 'Z':
		change_page(MT_PAGE_IDX_1JZVVTI);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef KNOCK_ALTERNATIVE
	  case 'F':
		change_page(MT_PAGE_IDX_KNOCK);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef BOOSTCONTROL
      case 'Y': // boost
		change_page(MT_PAGE_IDX_BOOST_REF);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef MAT_ENRICH
      case 'E': // temperatures
		change_page(MT_PAGE_IDX_TEMPERATURE);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef CHARGE_TEMP
      case 'E': // temperatures
		change_page(MT_PAGE_IDX_CHARGE);
		store_page(rx_offset, rx_data);
		break;
#endif

#ifdef MAP_CORR
	case 'I': // map_corr
		change_page(MT_PAGE_IDX_MAPCORR);
		store_page(rx_offset, rx_data);
		break;
#endif

	case 'W': // mat_retard
		change_page(MT_PAGE_IDX_MATRET);
		store_page(rx_offset, rx_data);
		break;

	case ']':
		change_page(MT_PAGE_IDX_LAUNCHRPM);
		store_page(rx_offset, rx_data);
		break;

    case SEND_PAGE_STUFFED_TYPE_COMMAND:{// sends page data in triggerlog format
            // rxdata stores the ID of the page if rx_offset ==0xff else ofset is the high address, rxdata is the low address
            uint8_t len;
            void* addr;
            if (rx_offset == 0xff){
                pushfuncpar(&SendPage_data_stuffed,0xFF,(void*)(0xff00|rx_data));//send the data(and rpmkpa) bytestuffed
                len = 0;
                addr = (void*)MEGATUNE_PAGE_POINTER(rx_data);
            }else{
                len = TRIGGERFRAME_READ_LEN_FIX;
                addr = (void*)( (rx_offset<<8) | rx_data);
            }
            pushfuncpar(&SendPage_stuffed,6+len,addr);// send the header
        }
        break;

      default:
        break;
    }
    com.comm_state = CMD; // back to normal mode
    break;

  default:
    switch (rx_data) {
    case 'A': // send all realtime values
	  aimdisable();
	  pushfunc((void*)MTsendRTvar, MT_RTVAR_COUNT);
      break;
    case 'B': // store to eeprom
	save_to_storage();
      break;

    case 'C': // test communications
      pushfunc(&MTtestComm, 1);
      break;
    case 'D': // send all realtime values
	  aimdisable();
	  pushfunc((void*)MTsendRTvar, MT_RTVAR_COUNTMORE+2);
      break;
    case READ_REALTIME_STUFFED_TYPE_COMMAND: // send all realtime values in triggerlog format
	  aimdisable();
	  pushfunc(MTsendRTvar_triggerlog, MT_RTVAR_COUNTMORE+4);
      break;
    case 'Q': // send code version
	  aimdisable();
      pushfunc(MTcodeVer, MT_CODEVER_LEN);
      break;

    case 'S': // send textual signature
	  aimdisable();
      pushfunc(&sendSignature, 26);
      break;

    case 'v': // send page
	  pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);  // don't know the length yet
      break;

	/* Separate command for each page write / read
	* No config/other overwrite on ecu reset. */

	// config read
	case 'o':
		aimdisable();
		change_page(MT_PAGE_IDX_CONFIG);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// config_ext read
	case 'g':
		change_page(MT_PAGE_IDX_CONFIGEXT);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// config_s read
	case 130://'�'
		change_page(MT_PAGE_IDX_CONFIG_SHARED);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// h[] table read
	case 'h':
		change_page(MT_PAGE_IDX_HWMAP);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// tps transfer and boosttable read
	case 't':
		change_page(MT_PAGE_IDX_BOOST);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// ve ref table read
	case 'j':
		change_page(MT_PAGE_IDX_VE);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// advance table read
	case 'n':
		change_page(MT_PAGE_IDX_IGNADV);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	// lambda table read
	case 'l':
		change_page(MT_PAGE_IDX_LAMBDAVIEW);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
	case '(':
		change_page(MT_PAGE_IDX_LAMBDA);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#ifdef A340E
	case 'u':
		change_page(MT_PAGE_IDX_A340E);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef TOYOTA_1JZ_VVTI
	case 'z':
		change_page(MT_PAGE_IDX_1JZVVTI);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef KNOCK_ALTERNATIVE
	case 'f':
		change_page(MT_PAGE_IDX_KNOCK);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

    // boost Ref table read
#ifdef BOOSTCONTROL
	case 'y':
		change_page(MT_PAGE_IDX_BOOST_REF);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef MAT_ENRICH
	case 'e':
		change_page(MT_PAGE_IDX_TEMPERATURE);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef CHARGE_TEMP
	case 'e':
		change_page(MT_PAGE_IDX_CHARGE);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

#ifdef MAP_CORR
	case 'i':
		change_page(MT_PAGE_IDX_MAPCORR);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;
#endif

	case 'w':
		change_page(MT_PAGE_IDX_MATRET);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

	case '[':
		change_page(MT_PAGE_IDX_LAUNCHRPM);
		pushfuncpar(&send_page, 0xff, (void *)(uint16_t)debug.megatune_page);
		break;

    case 'X': // forcing: channel + value
	case 'G': // config_ext write
	case 131: // config shared write
	case 'H': // h[] table write
	case 'T': // tps transfer and boosttable write
	case 'J': // ve ref table write
	case 'N': // advance table write
	case 'L': // lambda table write
	case ')': // raw lambda table write
	case 'E': // tempt table write
	case 'U': // TODO: ifdef A340E
	case 'F': // TODO: ifdef KNOCK_ALTERNATIVE
	case 'Z': // TODO: ifdef TOYOTA_1JZ_VVTI
	case 'Y': // boost Ref table write TODO: ifdef BOOSTCONTROL
	case 'W':
	case 'I':
	case 'O': // config write
	case ']': // launchrpm(wheelspeed) table
    case SEND_PAGE_STUFFED_TYPE_COMMAND: /* send the selected table in triggerlog format ( contains the address or pagename )*/
		com.comm_state = OFFSET;
		rx_cmd = rx_data;
		break;
    case GET_PAGE_LEN_CRC_STUFFED_TYPE_COMMAND: /* send the selected table in triggerlog format ( contains the address or pagename )*/
		pushfunc(&send_page_len_crc_stuffed,MT_PAGE_CRC_PACKET_LEN );
        break;

	default:
		break;
    }
  }
}

/* set needcrc to 2 if crc needed at the end of transfer
 * only makes sense for commands that send data to megatune, eg send_page()
 * (NOT for those that store)
 * Giant hack: we change the cmd_queue_head command to send commcrc after us
 */
static inline void terminate_transfer( uint8_t needcrc ) {
	cmd_queue[cmd_queue_head].count = 0xFF; // will become 0 in ringbuf_fill()
	cmd_queue[cmd_queue_head].len = needcrc ;
	cmd_queue[cmd_queue_head].func = (void*)0; // send 2 byte comm.crc
	cmd_queue[cmd_queue_head].private = (void*)(&com.commcrc);
}

/** send config and tables to vemstune */
uint8_t send_page(uint8_t i) {
	uint8_t val, size, page;

    uint16_t priv=(uint16_t)cmd_queue[cmd_queue_head].private;
	uint8_t needcrc=(priv>>8) ? 0 : 2; //set priv to zero if crc needed at the end of transfer
	page = (uint8_t)priv;
	size = MEGATUNE_PAGE_SIZE(page);

	if (page == MT_PAGE_IDX_DUMMY)
	    return(0);

	if( i < size ){
		uint8_t *tbl = MEGATUNE_PAGE_POINTER(page) ;
		val = tbl[i];

		if( page == MT_PAGE_IDX_LAMBDAVIEW ) {
		    return comm_getLambdaView(val);
		}

		if(page == MT_PAGE_IDX_BOOST) {
			// after t[] (with size TPSTRANSFERTABLE_SIZE) there comes
			// b[] (with size BOOSTCONTROL_GEARS * VE_SIZE_RPM):
			if( i >= TPSTRANSFERTABLE_SIZE ) {
				val = boost.boostmaxtarget_table[ i - TPSTRANSFERTABLE_SIZE ];
			}
		}

		if (!(page_has_rpmkpa(page)) && ( i >= size-1 )) {
			// don't need rpm/kpa headers?
			// so terminate
			terminate_transfer(needcrc);
		}

		return val;
	}

	val = 0; // suppress gcc warning

	struct bins_pointer_t bins_p;
	get_bins_pointer(page, &bins_p);

	i -= size;
	if (i < bins_p.rpm_size) {
		return bins_p.p_rpm_range[i];
	}
	i -= bins_p.rpm_size;
	if (i < bins_p.kpa_size )
		val = bins_p.p_kpa_range[i];
	if (i >= (bins_p.kpa_size - 1))
		terminate_transfer(needcrc);

	return val;
}

/**
 * some tables need special calculation
 */
void store_page(uint8_t addr, uint8_t value) {
	uint8_t size, page;

	page = debug.megatune_page;
	size = MEGATUNE_PAGE_SIZE(page);

	if (page == MT_PAGE_IDX_DUMMY)
        return;

	if (addr < size) {
		uint8_t *tbl;

		if(page == MT_PAGE_IDX_BOOST) {
			// after t[] there comes b[]:
			if( addr >= TPSTRANSFERTABLE_SIZE ){ // write to 2nd line
				boost.boostmaxtarget_table[ addr - TPSTRANSFERTABLE_SIZE ] = value;
				return;
			}
		}

		if(page == MT_PAGE_IDX_LAMBDAVIEW) {
		    value = comm_setLambdaView(value);
 		}

		// if( page == MT_PAGE_IDX_VE ) =>ve.ve8_table[addr] = value; //not special any longer
		tbl = MEGATUNE_PAGE_POINTER(page);
		tbl[addr] = value;
		return;
	}

	/*
	 * rpm_range and kpa_range arrays live at the end
	 */
	// need rpm/kpa headers?
	if(page_has_rpmkpa(page))
	{
		struct bins_pointer_t bins_p;
		get_bins_pointer(page, &bins_p);

		addr -= size;
		if ( addr < bins_p.rpm_size ) {
			bins_p.p_rpm_range[addr] = value;
			return;
		}
		addr -= bins_p.rpm_size;
		if (addr < bins_p.kpa_size )
			bins_p.p_kpa_range[addr] = value;
	}
}

void get_bins_pointer(uint8_t page, struct bins_pointer_t *bins_pointer)
{
	switch (page)
	{
#ifdef KNOCK_ALTERNATIVE
	case MT_PAGE_IDX_KNOCK:
		bins_pointer->rpm_size = KNOCK_SIZE_RPM;
		bins_pointer->kpa_size = KNOCK_SIZE_KPA;
		bins_pointer->p_rpm_range = knock_table.rpm_range;
		bins_pointer->p_kpa_range = knock_table.kpa_range;
		break;
#endif
	case MT_PAGE_IDX_BOOST_REF: // detached from ignadv bins since 1.1.63 (2009-11-11)
		bins_pointer->rpm_size = IGNADV_SIZE_RPM;
		bins_pointer->kpa_size = IGNADV_SIZE_KPA;
		bins_pointer->p_rpm_range = boost.boost_ref_rpm_range;
		bins_pointer->p_kpa_range = boost.boost_ref_kpa_range;
		break;
	case MT_PAGE_IDX_IGNADV:
	case MT_PAGE_IDX_1JZVVTI:
		// ign adv bins
		bins_pointer->rpm_size = IGNADV_SIZE_RPM;
		bins_pointer->kpa_size = IGNADV_SIZE_KPA;
		bins_pointer->p_rpm_range = ign.adv_table_rpm_range;
		bins_pointer->p_kpa_range = ign.adv_table_kpa_range;
		break;
	// all 8x8 could use: p_rpm_range= page+64 and p_kpa_range = page+72
	case MT_PAGE_IDX_LAMBDA:
	case MT_PAGE_IDX_LAMBDAVIEW:
		// lambda bins
		bins_pointer->rpm_size = LAMBDA_SIZE_RPM;
		bins_pointer->kpa_size = LAMBDA_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.lambda_rpm_range;
		bins_pointer->p_kpa_range = ve.lambda_kpa_range;
        break;
#ifdef MAP_CORR
	case MT_PAGE_IDX_MAPCORR:
		// map corr bins
		bins_pointer->rpm_size = MAP_CORR_SIZE_RPM;
		bins_pointer->kpa_size = MAP_CORR_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.map_corr_rpm_range;
		bins_pointer->p_kpa_range = ve.map_corr_kpa_range;
		break;
#endif
	case MT_PAGE_IDX_VE:
	case MT_PAGE_IDX_CHARGE:
		// ve bins
		bins_pointer->rpm_size = VE_SIZE_RPM;
		bins_pointer->kpa_size = VE_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.rpm_range;
		bins_pointer->p_kpa_range = ve.kpa_range;
		break;
	case MT_PAGE_IDX_MATRET:
		// lambda bins
	default:// cheat to prevent gcc varning
		bins_pointer->rpm_size = LAMBDA_SIZE_RPM;
		bins_pointer->kpa_size = LAMBDA_SIZE_KPA;
		bins_pointer->p_rpm_range = ve.mat_range;
		bins_pointer->p_kpa_range = ve.tps_range;
        break;
	}
}

/** because of byte stuffing, we push ourself. Min 2 bytes sent.
With some probability when i==1 we add 4 bytes: type(=1) + CRC + frame(=7E)
Called with 2 byte len automaticly close packet with 1/128 probability
*/
uint8_t TriggerlogPrimtrig( uint8_t i )
{
    com.primtrig_not_closed = 1;
	uint8_t v;
	uint16_t timestamp = (uint16_t)(cmd_queue[cmd_queue_head].private);

	if( i < 3 ){ // 0, 1 or 2
		if(i==0) v = timestamp >> 8; // send upper byte
		else if(i==1){
			v = timestamp & 0xFF; // send lower byte
			// with 1/128 probability extend length to 6 bytes:
			if( (v & 0x7F) == 0 ) cmd_queue[cmd_queue_head].len = 6;
		} else v = 1; // i ==2, type = 1: means CRC only
		ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
	} else { // 3 and 4 are CRC
		if( i == 5 ){
			ringbuf_push(&comm_sndbuf, HDLC_END); // send framing char, do NOT stuff
			com.commcrc = 0xffff;
			com.primtrig_not_closed = 0;
		} else { // 3 or 4
			v = (i == 3) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian
			ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
		}
	}
	return 'P'; // cheaper than return i, same as return 0
}
/** more than 6 bytes might be sent because of byte stuffing: we push ourself */
uint8_t TriggerlogAnytrig( uint8_t i )
{
	uint8_t v;
	uint16_t timestamp = (uint16_t)(cmd_queue[cmd_queue_head].private);
	if(i <= 2){
	  switch(i){
		case 0: v = timestamp >> 8; break; // send upper byte (Big endian)
		case 1: v = timestamp & 0xFF; break; // send lower byte
		default: v=cmd_queue[cmd_queue_head].privch.a; break; // type is from privch.a
	  }
	  ringbuf_push_bytestuff_updatecrc(&comm_sndbuf, v);
	} else if(i>=5){
		ringbuf_push(&comm_sndbuf, HDLC_END); // send framing char, do NOT stuff
		com.commcrc = 0xffff;
        com.primtrig_not_closed = 0;

	} else { // i == 3 or 4
		v = (i == 3) ? com.commcrc & 0xff : com.commcrc >> 8;// CRC sent in little endian as in SIPR
		ringbuf_push_bytestuff(&comm_sndbuf, v); // When sending crc itself => doNT update crc
	}
	return 'T';
}

/***************************************************************************/
/*** pushfunc                                                            ***/
/***************************************************************************/
uint8_t pushfunc(uint8_t (*f)(uint8_t i), uint8_t len) {
  uint8_t next_tail;

  next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return -1; // whoops, queue is full

  /* len starts at 0 */
//	if (len > 0) // this looks like a hack - it is! IT WAS!
//		len = len - 1;

  cmd_queue[cmd_queue_tail].func = f;
  cmd_queue[cmd_queue_tail].len = len;
  cmd_queue[cmd_queue_tail].count = 0;
  cmd_queue[cmd_queue_tail].private = 0;

  cmd_queue_tail = next_tail; // last step, when ALL other data is set

//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return 0;
}

/** Currently void, since cmd_queue overrun cannot be handled by caller. Caller could check BEFORE calling.
** same as pushfunc(), but more parameters
*/
void pushfuncpar2(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint8_t p2)
{
  uint8_t next_tail;

  next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return; // return -1; // whoops, queue is full

  cmd_queue[cmd_queue_tail].func = f;
  cmd_queue[cmd_queue_tail].len = len;
  cmd_queue[cmd_queue_tail].count = 0;
  cmd_queue[cmd_queue_tail].private = par;
  cmd_queue[cmd_queue_tail].privch.a = p2;
  cmd_queue_tail = next_tail;

//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return; // return 0;
}
//eg.: pushfuncpar2W(&srl_write_4byte_plus_end,7,(void*)((WRITE_EEPROM_STUFFED_TYPE_COMMAND - 128 ) | (eeprom.length <<8 )),(uint16_t)eeprom.p_eeprom);//send acknowledge on write (written cnt,address)
void pushfuncpar2W(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint16_t p2)
{
  uint8_t next_tail;

  next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return; // return -1; // whoops, queue is full

  cmd_queue[cmd_queue_tail].func = f;
  cmd_queue[cmd_queue_tail].len = len;
  cmd_queue[cmd_queue_tail].count = 0;
  cmd_queue[cmd_queue_tail].private = par;
  cmd_queue[cmd_queue_tail].privw = p2;
  cmd_queue_tail = next_tail;

//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return; // return 0;
}

// compatibility function with less parameters
void pushfuncpar(uint8_t (*f)(uint8_t i), uint8_t len, void * par)
{
	pushfuncpar2(f, len, par, 0);
}
/**
* Dumps a few tables, starting with wherever p and p+1 points to (travelling forward).
* makes sure entries < CMD_QUEUE_SIZE (not equal !!!)
*/
void pushfuncTableDump(uint8_t *p, uint8_t entries)
{
	// uint8_t allowed = CMD_QUEUE_SIZE-1; // not more than this many, please. Not needed if entries n< CMD_QUEUE_SIZE
	do {
		// pgm_read_word_near( p ); // endianess ?
		uint8_t table = pgm_read_byte_near( p++ );
		uint8_t lines = pgm_read_byte_near( p++ );
		pushfuncpar2(&TableDump, 0xFF, (void*)( (table << 8) + lines ), 0 );
	} while (--entries);
}
/* "small" wrapper because otherwise menu.c does not know size... Avoided because it bloated FLASH SIZE by +500 bytes !
void pushfuncTableDump1( void )
{
	 pushfuncTableDump( TABLES_TO_DUMP1, sizeof(TABLES_TO_DUMP1) >> 1 );
} */

/***************************************************************************/
/*** initUART                                                            ***/
/***************************************************************************/
void initUART(void) {
  /* reset the baudrate flag. 0-d by gcc at startup, but ... */
  new_baudrate=0;

  com.comm_state = CMD;

  /* rx enable, tx enable */
  //  UCSR0B = _BV(RXCIE0) | _BV(UDRIE0) | _BV(RXEN0) | _BV(TXEN0);
  UCSR0B = _BV(RXCIE0) | _BV(RXEN0) | _BV(TXEN0);
  /* 8 databit, 1 stopbit */
  UCSR0C = _BV(UCSZ01) | _BV(UCSZ00);

  /* 19200 baud @16MHz : 1000000 / (51+1) = 19231 */
  UBRR0H = 0;
  UBRR0L = 0x33;

  cmd_queue_tail = cmd_queue_head = 0;
	com.commcrc = 0xffff; // 12 bytes flash. Not needed for D command (set at offset 0), but useful for mde40 logging and possibly others in future
	com.primtrig_not_closed = 0;
}
/**
* initUART1 (1-wire)
*/
void initUART1(void)
{
#ifdef ONE_WIRE
    if ( config_s.prohibit_config & _BV(PROHIBIT_CONF_IBUTTON) ){// ibutton configured
     // rx enable
      //  UCSR0B = _BV(RXCIE0) | _BV(UDRIE0) | _BV(RXEN0) | _BV(TXEN0);
      UCSR1B = _BV(RXCIE1) | _BV(RXEN1) | _BV(TXEN1); // tx enable is needed
      /* 8 databit, 1 stopbit */
      UCSR1C = _BV(UCSZ11) | _BV(UCSZ10);
      UBRR1H = 0;      /* 9600 baud @16MHz */
      UBRR1L = 103;
      return;//GPSlog cannot work with 1wire
   }
#endif //ONE_WIRE
#ifdef GPSLOG
    if ( (sd_bytestuffed() )//GPS_LOG works only with bytestuffed SD log
        && ( (config.gps_log & _BV(GPS_LOG_ON_COM1_ENABLE_BIT)) == 0 )//GPS_LOG on first serial disabled => on second enabled
    ){
        // rx enable
        //  UCSR0B = _BV(RXCIE0) | _BV(UDRIE0) | _BV(RXEN0) | _BV(TXEN0);
        UCSR1B = _BV(RXCIE1) | _BV(RXEN1) | _BV(TXEN1); // tx enable is needed
        /* 8 databit, 1 stopbit */
        UCSR1C = _BV(UCSZ11) | _BV(UCSZ10);
        UBRR1H = 0;      /* 9600 baud @16MHz */
        UBRR1L = gps_baud();
        com.gps.state = GPS_DATA_EMPTY;
    }

#endif //GPSLOG
}

/***************************************************************************/
/*** INTERRUPT: UART rx complete ***/
/***************************************************************************/
// a small receivebuffer is still waiting to be implemented
SIGNAL(SIG_UART0_RECV) {

  // check status register for frame or overrun errors
  if ( UCSR0A & (_BV(FE0) | _BV(DOR0)) ) {
    uint8_t dummy;
    dummy = UDR0; // just read uart and throw away the result
    return;
  } else {
	ringbuf_push(&comm_recbuf, UDR0);
    // comm(UDR0);	// TODO: put to a buffer, which menu.c consumes and calls back comm() if necessary
  }
}

#ifdef ONE_WIRE
void one_wire_rx(uint8_t readchar)
{
    uint8_t rx_cnt;
  	rx_cnt = one_wire.rx_cnt;
  	rx_cnt++;
  	if ((rx_cnt == 2) && (readchar == 0x33)) one_wire.rx_crc=0;
  	else{
  	    if (rx_cnt == 10){//crc incoming
  	        if (one_wire.rx_crc == readchar) one_wire.status = ONE_WIRE_CRC_OK;
  	        else one_wire.status = ONE_WIRE_CRC_FAIL;
  	    }else{
            if (rx_cnt == 4)
                one_wire.msB=readchar;
            else if (rx_cnt == 5)
                one_wire.ssB=readchar;
            one_wire.rx_crc = _crc_ibutton_update(one_wire.rx_crc,readchar);
  	    }
  	}
  	one_wire.rx_cnt = rx_cnt;

}
#endif //ONE_WIRE

/**
* INTERRUPT: UART1 (1-wire) rx complete or second GPS input
*/
SIGNAL(SIG_UART1_RECV) {
	uint8_t readchar;
	readchar = UDR1;	// read uart
	// check status register for frame or overrun errors
	if ( UCSR1A & (_BV(FE1) | _BV(DOR1)) ) {
		 // just and throw away the result
    		return;
	}
	if( com.comm1_relay_allowed ){
	// inject into char-buf of UART0 directly
	// BEWARE: possibly into the middle of some bigger block, so only useful for testing
		ringbuf_push(&comm_sndbuf, readchar);
		UCSR0B |= _BV(UDRIE0); // enable transmit interrupt for UART0, so comm_sndbuf does not get stuck
  	}
#ifdef ONE_WIRE
    if ( config_s.prohibit_config & _BV(PROHIBIT_CONF_IBUTTON) ){// ibutton configured
	    one_wire_rx(readchar);
	    return;//GPSlog cannot work with 1wire
	}
#endif //ONE_WIRE
#ifdef GPSLOG
    if (sd_bytestuffed()){//GPS_LOG works only with bytestuffed SD log
        gps_parse(readchar);
    }
#endif
}

#ifdef ONE_WIRE
/**
* INTERRUPT: UART1 (1-wire) tx empty
*/
SIGNAL(SIG_UART1_DATA) {
    uint8_t tx_cnt = one_wire.tx_cnt;
    UDR1 = pgm_read_byte( &(oneWireInit[tx_cnt++] ));//send the next byte
    if (tx_cnt == oneWireInitLen){//one_wire init finished
        UCSR1B &=~_BV(UDRE1);// disable data register empty interrupt on usart1
        tx_cnt = 0;//reenable 1wire init
    }
    one_wire.tx_cnt = tx_cnt;
}
#endif //ONE_WIRE
/***************************************************************************/
/*** INTERRUPT: UART tx empty ***/
/***************************************************************************/
SIGNAL(SIG_UART0_DATA) {
	uint16_t d;

	d = ringbuf_shift(&comm_sndbuf);

	if (d>>8) { // buffer empty
                // this is tricky. To save on useless processing of the SIG_UART0_TRANS
                // interrupt we will enable that interrupt to signal us when this last byte
                // has been sent if and only if there is a baudrate change pending
		UCSR0B &= ~_BV(UDRIE0);    // disable TX interrupts
                if (new_baudrate!=0)
		  UCSR0B |= _BV(TXCIE0);    // enable TX complete interrupt
	} else {
		UDR0 = d;
	}

}


/***************************************************************************/
/*** INTERRUPT: UART tx complete ***/
/***************************************************************************/
SIGNAL(SIG_UART0_TRANS) {
  // we can't change the baudrate when we get a TX empty interrupt because
  // the shift register still contains the last byte to be sent. So here we
  // check to make sure the TX interrupts are disabled (nothing buffered)
  // and if the new_baudrate flag is set then we're good to change the baudrate
  // furthermore since we don't want to waste any CPU cycles we will disable
  // our interrupt.

  // if the TX interrupt is enabled then someone must have sent data while
  // we were waiting for the last byte to send and refilled the buffer so
  // disable our interrupt and bail
  if (UCSR0B & _BV(UDRIE0)) {
    UCSR0B &= ~_BV(TXCIE0);    // disable the TX complete interrupt
  }
  else {
    // if there is a baudrate change pending then make sure the cmd_queue is empty
    // and change the baudrate. Otherwise disable TX interrupts
    if ( (new_baudrate!=0) && (cmd_queue_tail==cmd_queue_head)) {
      change_baudrate(new_baudrate);
      new_baudrate=0;
      UCSR0B &= ~_BV(TXCIE0);    // disable the TX complete interrupt
    }
  }
}

// Use for debugging
void put_char(uint8_t byte)
{
    while( !(UCSR0A & 0x20) )
        ;
    UDR0 = byte;
}

uint8_t comm_getLambdaView(uint8_t val)
{
    uint16_t mt_val;
    mt_val = mult16_8((val + LAMBDA_INT_OFFSET), 72) >> 7;
    mt_val = (mult16_8(div_65535_x(mt_val), 72) >> 7) - MT_LAMBDA_OFFSET;
    return mt_val & 0xFF;
}

uint8_t comm_setLambdaView(uint8_t val)
{
    uint16_t int_val;
    int_val = mult16_8((val + MT_LAMBDA_OFFSET), 72) >> 7;
    int_val = mult16_8(div_65535_x(int_val) , 72) >> 7;
    val = (int_val - LAMBDA_INT_OFFSET) & 0xFF;
    return val;
}
